import datetime
import logging
from typing import List

import sqlalchemy
from airflow import DAG, settings
from airflow.models import DagModel
from airflow.utils.session import provide_session
from deprecated import deprecated
from sqlalchemy import MetaData, Table
from sqlalchemy.exc import NoSuchTableError
from sqlalchemy.orm import Session
from airflow.models import Connection
from airflow.models import Pool


@provide_session
def get_connections(session: Session) -> List["Connection"]:
    connections = session.query(Connection).order_by(Connection.conn_id).all()
    return connections


@provide_session
def get_pools(session: Session) -> List["Pool"]:
    pools = session.query(Pool).all()
    return pools


def get_pool(pool_name):
    return [p for p in get_pools() if p.pool == pool_name][0]


@provide_session
def get_variables(session: Session):
    from airflow.models import Variable

    variables = session.query(Variable).order_by(Variable.key).all()
    return variables


def get_variable(variable: str):
    return [v for v in get_variables() if v.key == variable][0]


def get_dags():
    from airflow.models import DagBag

    dags = DagBag().dags
    return dags


def get_dag(dag_id: str) -> DAG:
    return get_dags()[dag_id]


def set_dag_is_paused(dag_id, is_paused):
    DagModel.get_dagmodel(dag_id).set_is_paused(is_paused=is_paused)


@deprecated
class LocalAirflowClient:
    pass


@provide_session
def migrate(session: Session, table_name: str, dag_id: str):
    logging.info("Creating source SQL Connections ..")
    source_engine = session.get_bind()
    source_metadata_obj = MetaData(bind=source_engine)
    source_table = get_table(source_metadata_obj, source_engine, table_name)

    # noinspection SqlInjection
    logging.info(f"Fetching data for {table_name} from source...")
    source_result = source_engine.execution_options(stream_results=True).execute(
        f"SELECT * FROM {source_table.name} where dag_id='{dag_id}' order by start_date desc limit 5"
    )
    rtn = []
    for result in source_result.fetchall():
        result = dict(result._asdict())
        for key, value in result.items():
            if isinstance(value, memoryview):
                result[key] = value.tobytes().decode("iso-8859-1")
            if isinstance(value, datetime.datetime):
                result[key] = f"{value.strftime('%Y-%m-%d %H:%M:%S.%f+00')}"
        result["table"] = table_name
        rtn.append(result)
        source_table = get_table(source_metadata_obj, source_engine, "task_instance")
        sub_result = source_engine.execution_options(stream_results=True).execute(
            f"SELECT * FROM {source_table.name} where dag_id='{dag_id}' and run_id='{result['run_id']}'"
        )
        for ti_result in sub_result.fetchall():
            ti_result_new = dict(ti_result._asdict())
            ti_result = dict(ti_result._asdict())
            ti_result["table"] = "task_instance"
            for ti_key, ti_value in ti_result_new.items():
                if isinstance(ti_value, memoryview):
                    del ti_result[ti_key]
                if isinstance(ti_value, datetime.datetime):
                    ti_result[
                        ti_key
                    ] = f"{ti_value.strftime('%Y-%m-%d %H:%M:%S.%f+00')}"
            rtn.append(ti_result)

    return rtn


def receive_dag(data: list = []):
    dest_session = settings.Session()
    dest_engine = dest_session.get_bind()
    dest_metadata_obj = MetaData(bind=dest_engine)
    for datum in data:
        try:
            del datum["conf"]
            del datum["id"]
        except Exception:
            logging.info(
                'tried to deleting something that doesn"t exist, don"t worry it still doesn"t exist'
            )
        table_name = datum.pop("table")
        dest_table = get_table(dest_metadata_obj, dest_engine, table_name)
        dest_engine.execute(
            sqlalchemy.dialects.postgresql.insert(dest_table).on_conflict_do_nothing(),
            [datum],
        )
        dest_session.commit()


def get_table(_metadata_obj, _engine, _table_name):
    try:
        return Table(f"airflow.{_table_name}", _metadata_obj, autoload_with=_engine)
    except NoSuchTableError:
        return Table(_table_name, _metadata_obj, autoload_with=_engine)



import json
from typing import Any, List

import requests
from airflow.models import Connection, Pool, Variable
from cachetools.func import ttl_cache
from deprecated import deprecated
from requests import Response

from astronomer.starship.services import local_airflow_client


def conn_to_json(connection: Connection) -> dict:
    return {
        "connection_id": connection.conn_id,
        "conn_type": connection.conn_type,
        "host": connection.host,
        "login": connection.login,
        "schema": connection.schema,
        "port": connection.port,
        "password": connection.password or "",
        "extra": connection.extra,
    }


@ttl_cache(ttl=1)
def get_connections(deployment_url: str, token: str) -> List[Any]:
    r = requests.get(
        f"{deployment_url}/api/v1/connections",
        headers={"Authorization": f"Bearer {token}"},
    )
    r.raise_for_status()
    return r.json()["connections"]


def delete_connection(
    deployment_url: str, token: str, connection: Connection
) -> Response:
    r = requests.delete(
        f"{deployment_url}/api/v1/connections/{connection.conn_id}",
        headers={"Authorization": f"Bearer {token}"},
    )
    r.raise_for_status()
    return r


def do_test_connection(
    deployment_url: str, token: str, connection: Connection
) -> Response:
    r = requests.post(
        f"{deployment_url}/api/v1/connections/test",
        headers={"Authorization": f"Bearer {token}"},
        json=conn_to_json(connection),
    )
    r.raise_for_status()
    return r


def create_connection(deployment_url, token, connection) -> Response:
    r = requests.post(
        f"{deployment_url}/api/v1/connections",
        headers={"Authorization": f"Bearer {token}"},
        json=conn_to_json(connection),
    )
    r.raise_for_status()
    return r


def delete_pool(deployment_url, token, pool) -> Response:
    r = requests.delete(
        f"{deployment_url}/api/v1/pools/{pool.pool}",
        headers={"Authorization": f"Bearer {token}"},
    )
    r.raise_for_status()
    return r


def get_pools(deployment_url: str, token: str):
    r = requests.get(
        f"{deployment_url}/api/v1/pools",
        headers={"Authorization": f"Bearer {token}"},
    )
    r.raise_for_status()
    return r.json()["pools"]


def create_pool(deployment_url, token, pool: Pool) -> Response:
    r = requests.post(
        f"{deployment_url}/api/v1/pools",
        headers={"Authorization": f"Bearer {token}"},
        json={"name": pool.pool, "slots": pool.slots, "description": pool.description},
    )
    r.raise_for_status()
    return r


def is_pool_migrated(deployment_url: str, token: str, pool_name: str):
    remote_pools = get_pools(deployment_url, token)
    return pool_name in (remote_pool.get("name", "") for remote_pool in remote_pools)


def get_variables(deployment_url: str, token: str):
    r = requests.get(
        f"{deployment_url}/api/v1/variables",
        headers={"Authorization": f"Bearer {token}"},
    )
    r.raise_for_status()
    return r.json()["variables"]


def delete_variable(deployment_url: str, token: str, variable: Variable):
    r = requests.delete(
        f"{deployment_url}/api/v1/variables/{variable.key}",
        headers={"Authorization": f"Bearer {token}"},
    )
    r.raise_for_status()
    return r


def is_variable_migrated(deployment_url: str, token: str, variable: str):
    return variable in (v["key"] for v in get_variables(deployment_url, token))


def create_variable(deployment_url, token: str, variable: Variable):
    r = requests.post(
        f"{deployment_url}/api/v1/variables",
        headers={"Authorization": f"Bearer {token}"},
        json={"key": variable.key, "value": variable.val},
    )
    r.raise_for_status()
    return r


@deprecated(reason="unused, doesn't work in astro")
def get_config(deployment_url: str, token: str):
    r = requests.get(
        f"{deployment_url}/api/v1/config",
        headers={"Authorization": f"Bearer {token}"},
    )
    r.raise_for_status()
    return r.json()


def set_dag_is_paused(dag_id, is_paused, deployment_url, token):
    r = requests.patch(
        f"{deployment_url}/api/v1/dags?dag_id_pattern={dag_id}",
        headers={"Authorization": f"Bearer {token}"},
        json={"is_paused": is_paused},
    )
    r.raise_for_status()
    return r


def get_dag(dag_id, deployment_url, token) -> Response:
    r = requests.get(
        f"{deployment_url}/api/v1/dags/{dag_id}",
        headers={"Authorization": f"Bearer {token}"},
    )
    r.raise_for_status()
    return r


def get_dag_runs(dag_id, deployment_url, token) -> Response:
    r = requests.get(
        f"{deployment_url}/api/v1/dags/{dag_id}/dagRuns",
        headers={"Authorization": f"Bearer {token}"},
    )
    r.raise_for_status()
    return r


def migrate_dag(dag: str, deployment_url: str, token: str):
    result = local_airflow_client.migrate(table_name="dag_run", dag_id=dag)
    r = requests.post(
        f"{deployment_url}/astromigration/dag_history/receive",
        data=json.dumps(result),
        headers={
            "Content-Type": "application/json",
            "Authorization": f"Bearer {token}",
        },
    )
    r.raise_for_status()
    return r.ok



import os
import logging
from typing import Optional
import jwt
from airflow.plugins_manager import AirflowPlugin
from airflow.security import permissions
from airflow.www import auth
from airflow.www.app import csrf
from flask import Blueprint, Response, redirect, request, session, url_for
from flask_appbuilder import BaseView as AppBuilderBaseView
from flask_appbuilder import expose

from astronomer.starship.services import (
    astro_client,
    local_airflow_client,
    remote_airflow_client,
)
from astronomer.starship.services.astro_client import is_environment_variable_migrated
from astronomer.starship.services.remote_airflow_client import is_pool_migrated

bp = Blueprint(
    "starship",
    __name__,
    template_folder="templates",
    static_folder="static",
    static_url_path="/static/starship",
)


def get_page_data(page):
    return {
        "AstroMigration.tabs_vars": {
            "component": "variables",
            "vars": {var.key: var for var in local_airflow_client.get_variables()},
        },
        "AstroMigration.tabs_dags": {
            "component": "dags",
            "dags": local_airflow_client.get_dags(),
        },
        "AstroMigration.tabs_pools": {
            "component": "pools",
            "pools": {pool.pool: pool for pool in local_airflow_client.get_pools()},
        },
        "AstroMigration.tabs_conns": {
            "component": "connections",
            "conns": {
                conn.conn_id: conn for conn in local_airflow_client.get_connections()
            },
        },
        "AstroMigration.tabs_env": {
            "component": "env",
            "environ": os.environ,
        },
    }[page]


class AstroMigration(AppBuilderBaseView):
    default_view = "main"

    @expose("/", methods=["GET", "POST"])
    @auth.has_access([(permissions.ACTION_CAN_READ, permissions.RESOURCE_CONFIG)])
    def main(self):
        session.update(request.form)
        return self.render_template(
            "starship/main.html",
            data={"tab": session.get("tab", "AstroMigration.tabs_conns")},
        )

    @expose("/dag_history/receive", methods=["GET", "POST"])
    @csrf.exempt
    def receive_dag_history(self):
        data = request.json
        try:
            local_airflow_client.receive_dag(data=data)
            return Response("OK", 200)
        except Exception as e:
            logging.exception(e)
            return Response(str(e), 200)

    @expose("/modal/token")
    def modal_token_entry(self):
        token = session.get("token")

        if not token:
            return self.render_template(
                "starship/components/token_modal.html", show=True, error=None
            )

        try:
            jwt.decode(
                token,
                audience=["astronomer-ee"],
                key=astro_client.get_jwk(token).key,
                algorithms=["RS256"],
            )
        except jwt.exceptions.ExpiredSignatureError:
            return self.render_template(
                "starship/components/token_modal.html", show=True, error="expired"
            )
        except jwt.exceptions.InvalidTokenError:
            return self.render_template(
                "starship/components/token_modal.html", show=True, error="invalid"
            )

        return self.render_template("starship/components/token_modal.html", show=False)

    @expose("/button/save_token", methods=("POST",))
    @auth.has_access([(permissions.ACTION_CAN_READ, permissions.RESOURCE_CONFIG)])
    def button_save_token(self):
        session["token"] = request.form.get("astroUserToken")
        tab = session.get("tab", "AstroMigration.tabs_conns")
        return self.render_template("starship/migration.html", data={"tab": tab})

    @expose("/tabs/dags")
    @auth.has_access([(permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG)])
    def tabs_dags(self):
        session["tab"] = "AstroMigration.tabs_dags"
        return self.render_template(
            "starship/dags.html", data=get_page_data(session["tab"])
        )

    @expose("/tabs/variables")
    @auth.has_access([(permissions.ACTION_CAN_READ, permissions.RESOURCE_VARIABLE)])
    def tabs_vars(self):
        session["tab"] = "AstroMigration.tabs_vars"
        return self.render_template(
            "starship/variables.html", data=get_page_data(session["tab"])
        )

    @expose("/tabs/pools")
    @auth.has_access([(permissions.ACTION_CAN_READ, permissions.RESOURCE_VARIABLE)])
    def tabs_pools(self):
        session["tab"] = "AstroMigration.tabs_pools"
        return self.render_template(
            "starship/pools.html", data=get_page_data(session["tab"])
        )

    @expose("/tabs/connections")
    @auth.has_access([(permissions.ACTION_CAN_READ, permissions.RESOURCE_CONNECTION)])
    def tabs_conns(self):
        session["tab"] = "AstroMigration.tabs_conns"
        return self.render_template(
            "starship/connections.html", data=get_page_data(session["tab"])
        )

    @expose("/tabs/env")
    @auth.has_access([(permissions.ACTION_CAN_READ, permissions.RESOURCE_CONFIG)])
    def tabs_env(self):
        session["tab"] = "AstroMigration.tabs_env"
        return self.render_template(
            "starship/env.html", data=get_page_data(session["tab"])
        )

    @expose(
        "/button_migrate_connection/<string:deployment>/<string:conn_id>",
        methods=("GET", "POST"),
    )
    @auth.has_access([(permissions.ACTION_CAN_READ, permissions.RESOURCE_CONNECTION)])
    def button_migrate_connection(self, conn_id: str, deployment: str):
        token = session.get("token")
        deployment_url = astro_client.get_deployment_url(deployment, token)

        if request.method == "POST":
            local_connections = {
                conn.conn_id: conn for conn in local_airflow_client.get_connections()
            }
            remote_airflow_client.create_connection(
                deployment_url, token, local_connections[conn_id]
            )

        deployment_conns = remote_airflow_client.get_connections(deployment_url, token)

        is_migrated = conn_id in [
            remote_conn["connection_id"] for remote_conn in deployment_conns
        ]

        return self.render_template(
            "starship/components/migrate_connection_button.html",
            conn_id=conn_id,
            deployment=deployment,
            is_migrated=is_migrated,
        )

    @expose("/label_test_connection/<string:deployment>/<string:conn_id>")
    @auth.has_access([(permissions.ACTION_CAN_READ, permissions.RESOURCE_CONNECTION)])
    def label_test_connection(self, conn_id: str, deployment: str):
        token = session.get("token")
        deployment_url = astro_client.get_deployment_url(deployment, token)

        local_connection = {
            conn.conn_id: conn for conn in local_airflow_client.get_connections()
        }[conn_id]
        r = remote_airflow_client.do_test_connection(
            deployment_url, token, local_connection
        )

        return self.render_template(
            "starship/components/test_connection_label.html",
            data=r.json(),
            conn_id=conn_id,
        )

    @expose(
        "/button/migrate/var/<string:deployment>/<string:variable>",
        methods=("GET", "POST"),
    )
    @auth.has_access([(permissions.ACTION_CAN_READ, permissions.RESOURCE_VARIABLE)])
    def button_migrate_variable(self, variable: str, deployment: str):
        token = session.get("token")
        deployment_url = astro_client.get_deployment_url(deployment, token)

        if request.method == "POST":
            remote_airflow_client.create_variable(
                deployment_url, token, local_airflow_client.get_variable(variable)
            )

        is_migrated = remote_airflow_client.is_variable_migrated(
            deployment_url, token, variable
        )

        return self.render_template(
            "starship/components/migrate_variable_button.html",
            variable=variable,
            deployment=deployment,
            is_migrated=is_migrated,
        )

    @expose(
        "/button/migrate/pool/<string:deployment>/<string:pool_name>",
        methods=("GET", "POST"),
    )
    @auth.has_access([(permissions.ACTION_CAN_READ, permissions.RESOURCE_VARIABLE)])
    def button_migrate_pool(self, pool_name: str, deployment: Optional[str] = None):
        token = session.get("token")
        if deployment:
            deployment_url = astro_client.get_deployment_url(deployment, token)

            if request.method == "POST":
                pool = local_airflow_client.get_pool(pool_name)
                remote_airflow_client.create_pool(deployment_url, token, pool)

            is_migrated = is_pool_migrated(deployment_url, token, pool_name)
        else:
            # Deployment hasn't been selected yet
            is_migrated = False

        return self.render_template(
            "starship/components/migrate_pool_button.html",
            pool=pool_name,
            deployment=deployment,
            is_migrated=is_migrated,
        )

    @expose("/button/migrate/env/<string:deployment>", methods=("POST",))
    @auth.has_access([(permissions.ACTION_CAN_READ, permissions.RESOURCE_CONFIG)])
    def button_migrate_env(self, deployment: str):
        token = session.get("token")
        items = dict(**request.form)
        del items["csrf_token"]
        astro_client.set_changed_environment_variables(
            deployment, token, iter(items.values())
        )
        return self.tabs_env()

    @expose("/checkbox/migrate/env/<string:deployment>/<string:key>/", methods=("GET",))
    @auth.has_access([(permissions.ACTION_CAN_READ, permissions.RESOURCE_CONFIG)])
    def checkbox_migrate_env(self, key: str, deployment: str):
        token = session.get("token")
        is_migrated = is_environment_variable_migrated(deployment, token, key)
        return self.render_template(
            "starship/components/env_checkbox.html",
            target=key,
            deployment=deployment,
            is_migrated=is_migrated,
        )

    @expose("/component/astro-deployment-selector")
    def deployment_selector(self):
        deployments = session.get("deployments")
        if not deployments:
            deployments = astro_client.get_deployments(session.get("token"))
            session["deployments"] = deployments

        user = session.get("user")
        if not user:
            user = astro_client.get_username(token=session.get("token"))
            session["user"] = user

        return self.render_template(
            "starship/components/target_deployment_select.html",
            deployments=deployments,
            username=user,
        )

    @expose("/logout")
    def logout(self):
        session.pop("token")
        session.pop("user")
        session.pop("deployments")
        return redirect(url_for("Airflow.index"))

    @expose("/component/row/dags/<string:deployment>/<string:dag_id>")
    @auth.has_access([(permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG)])
    def dag_cutover_row_get(self, deployment: str, dag_id: str):
        return self.dag_cutover_row(deployment, dag_id)

    @expose(
        "/component/row/dags/<string:deployment>/<string:dest>/<string:dag_id>/<string:action>",
        methods=(
            "GET",
            "POST",
        ),
    )
    @auth.has_access([(permissions.ACTION_CAN_EDIT, permissions.RESOURCE_DAG)])
    def dag_cutover_row(
        self, deployment: str, dag_id: str, dest: str = "local", action: str = "init"
    ):
        if dest not in ["local", "astro"]:
            raise RuntimeError("dest must be 'local' or 'astro'")
        if action not in ["pause", "unpause", "migrate", "init"]:
            raise Exception("action must be 'pause', 'unpause', or 'migrate'")

        token = session.get("token")
        deployment_url = astro_client.get_deployment_url(deployment, token)

        if request.method == "POST":
            if action == "migrate":
                remote_airflow_client.migrate_dag(
                    dag_id, deployment_url=deployment_url, token=token
                )
            else:
                is_paused = "pause" == action
                if dest == "local":
                    local_airflow_client.set_dag_is_paused(dag_id, is_paused)
                else:
                    remote_airflow_client.set_dag_is_paused(
                        dag_id, is_paused, deployment_url, token
                    )

        r = remote_airflow_client.get_dag(dag_id, deployment_url, token)
        dag_runs = remote_airflow_client.get_dag_runs(
            dag_id, deployment_url, token
        ).json()
        is_on_astro = not r.status_code == 404
        remote_dag = r.json()
        local_dag = local_airflow_client.get_dag(dag_id)

        return self.render_template(
            "starship/components/dag_row.html",
            dag_={
                "id": local_dag.dag_id,
                "is_on_astro": is_on_astro,
                "is_paused_here": local_dag.is_paused,
                "is_paused_on_astro": remote_dag["is_paused"] if is_on_astro else False,
                "has_history_on_astro": bool(dag_runs["total_entries"]),
            },
        )


v_appbuilder_view = AstroMigration()

v_appbuilder_package = {
    "name": "Migration Tool 🚀 Starship",
    "category": "Astronomer",
    "view": v_appbuilder_view,
}


class StarshipPlugin(AirflowPlugin):
    name = "starship"
    flask_blueprints = [bp]
    appbuilder_views = [v_appbuilder_package]



import requests

from airflow.models import BaseOperator, Variable

from airflow.utils.session import provide_session
from sqlalchemy.orm import Session
from typing import Any, Sequence


class AstroVariableMigrationOperator(BaseOperator):
    """
    Sends variables from Airflow metadatabase to Astronomer Deployment
    """

    template_fields: Sequence[str] = ("token", "deployment_url")
    ui_color = "#974bde"

    def __init__(
        self, deployment_url, token, variable_exclude_list=None, **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_url = deployment_url
        self.token = token
        self.variable_exclude_list = variable_exclude_list

    @provide_session
    def execute(self, context: Any, session: Session) -> None:
        get_variables = session.query(Variable).all()
        local_vars = {var.key: var.val for var in get_variables}

        if self.variable_exclude_list:
            exclude_list = self.variable_exclude_list
        else:
            exclude_list = []

        for key, value in local_vars.items():
            if key not in exclude_list:
                requests.post(
                    url=f"{self.deployment_url}/api/v1/variables",
                    headers={"Authorization": f"Bearer {self.token}"},
                    json={"key": key, "value": value},
                )



from typing import Any, Dict, List, Union

import base64
import json
import logging
from json import JSONDecodeError


def clean_airflow_report_output(log_string: str) -> Union[dict, str]:
    r"""Look for the magic string from the Airflow report and then decode the base64 and convert to json
    Or return output as a list, trimmed and split on newlines
    >>> clean_airflow_report_output('INFO 123 - xyz - abc\n\n\nERROR - 1234\n%%%%%%%\naGVsbG8gd29ybGQ=')
    'hello world'
    >>> clean_airflow_report_output('INFO 123-xyz - abc\n\nERROR-1234\n%%%%%%%\neyJvdXRwdXQiOiAiaGVsbG8gd29ybGQifQ==')
    {'output': 'hello world'}
    """

    def get_json_or_clean_str(o: str) -> Union[List[Any], Dict[Any, Any], Any]:
        """Either load JSON (if we can) or strip and split the string, while logging the error"""
        try:
            return json.loads(o)
        except (JSONDecodeError, TypeError) as e:
            logging.debug(e)
            logging.debug(o)
            return o.strip()

    log_lines = log_string.split("\n")
    enumerated_log_lines = list(enumerate(log_lines))
    found_i = -1
    for i, line in enumerated_log_lines:
        if "%%%%%%%" in line:
            found_i = i + 1
            break
    if found_i != -1:
        output = base64.decodebytes(
            "\n".join(log_lines[found_i:]).encode("utf-8")
        ).decode("utf-8")
        try:
            return json.loads(output)
        except JSONDecodeError:
            return get_json_or_clean_str(output)
    else:
        return get_json_or_clean_str(log_string)



import datetime
import json
import os
import socket
import urllib.error
from contextlib import redirect_stderr, redirect_stdout
from textwrap import dedent

import io
import runpy
from urllib.request import urlretrieve
import requests
from airflow.plugins_manager import AirflowPlugin
from flask import Blueprint, Response, flash, request
from flask_appbuilder import BaseView as AppBuilderBaseView
from flask_appbuilder import expose
from wtforms import Form, StringField, validators

from astronomer.aeroscope.util import clean_airflow_report_output

bp = Blueprint(
    "starship_aeroscope",
    __name__,
    template_folder="templates",  # registers airflow/plugins/templates as a Jinja template folder
    static_folder="static",
    static_url_path="/static/aeroscope",
)


class AeroForm(Form):
    organization = StringField("Organization", [validators.Length(min=4, max=25)])
    presigned_url = StringField(
        "Pre-signed URL (optional)",
        description=dedent(
            """The pre-signed URL field is optional and is to be supplied by an Astronomer Representative.
    If the pre-signed URL is used, the results of the Telescope Report is shipped to Astronomer."""
        ),
        validators=[validators.URL(), validators.optional()],
    )


# Creating a flask appbuilder BaseView
class StarshipAeroscope(AppBuilderBaseView):
    default_view = "aeroscope"

    @expose("/", methods=("GET", "POST"))
    def aeroscope(self):
        form = AeroForm(request.form)
        if (
            request.method == "POST"
            and form.validate()
            and request.form["action"] == "Send/Download Report"
        ):
            VERSION = os.getenv("TELESCOPE_REPORT_RELEASE_VERSION", "latest")
            a = "airflow_report.pyz"
            if VERSION == "latest":
                urlretrieve(
                    "https://github.com/astronomer/telescope/releases/latest/download/airflow_report.pyz",
                    a,
                )
            else:
                try:
                    urlretrieve(
                        f"https://github.com/astronomer/telescope/releases/download/{VERSION}/airflow_report.pyz",
                        a,
                    )
                except urllib.error.HTTPError as e:
                    flash(
                        f"Error finding specified version:{VERSION} -- Reason:{e.reason}"
                    )
            s = io.StringIO()
            with redirect_stdout(s), redirect_stderr(s):
                runpy.run_path(a)
            date = datetime.datetime.now(datetime.timezone.utc).isoformat()[:10]
            content = {
                "telescope_version": "aeroscope-latest",
                "report_date": date,
                "organization_name": form.organization.data,
                "local": {
                    socket.gethostname(): {
                        "airflow_report": clean_airflow_report_output(s.getvalue())
                    }
                },
            }
            if len(form.presigned_url.data) > 1:
                upload = requests.put(form.presigned_url.data, data=json.dumps(content))
                if upload.ok:
                    flash("Upload successful")
                else:
                    flash(upload.reason, "error")
            filename = f"{date}.{form.organization.data}.data.json"
            return Response(
                json.dumps(content),
                mimetype="application/json",
                headers={"Content-Disposition": f"attachment;filename={filename}"},
            )
        else:
            return self.render_template("aeroscope/main.html", form=form)


v_appbuilder_view = StarshipAeroscope()


# Defining the plugin class
class AeroscopePlugin(AirflowPlugin):
    name = "starship_aeroscope"
    hooks = []
    macros = []
    flask_blueprints = [bp]
    appbuilder_views = [
        {
            "name": "Run Report 🔭 Telescope",
            "category": "Astronomer",
            "view": v_appbuilder_view,
        },
    ]
    appbuilder_menu_items = []
    global_operator_extra_links = []
    operator_extra_links = []



import logging
import os
from typing import Optional, Iterator, Any, Dict
from urllib.parse import urlparse

import jwt
import requests
from cachetools.func import ttl_cache
from pydash import at
from python_graphql_client import GraphqlClient
from deprecated import deprecated

ASTRO_AUTH = os.environ.get("STARSHIP_ASTRO_AUTH", "https://auth.astronomer.io")
ASTROHUB_API = os.environ.get(
    "STARSHIP_ASTROHUB_API", "https://api.astronomer.io/hub/v1"
)
ASTROHUB_GRAPHQL_API = os.environ.get(
    "STARSHIP_ASTROHUB_GRAPHQL_API", "https://api.astronomer.io/hub/graphql"
)
ASTRO_ALPHA_API = os.environ.get(
    "STARSHIP_ASTRO_ALPHA_API", "https://api.astronomer.io/v1alpha1"
)


def get_username(token):
    headers = {"Authorization": f"Bearer {token}"}
    client = GraphqlClient(endpoint=ASTROHUB_API, headers=headers)
    query = "{self {user {username}}}"

    try:
        api_rv = at(client.execute(query), "data.self.user.username")[0]
        return api_rv
    except Exception as e:
        print(e)
        return None


@ttl_cache(ttl=1)
def get_deployments(token):
    if not token:
        return {}
    headers = {"Authorization": f"Bearer {token}"}
    orgs = get_organizations(token)
    short_name = os.getenv(
        "STARSHIP_ORG_SHORTNAME", [_ for _ in orgs.values()][0]["shortName"]
    )

    # FIXME use the first org for now. change to a select in the near term.
    url = f"{ASTRO_ALPHA_API}/organizations/{short_name}/deployments"

    try:
        r = requests.get(url, headers=headers)
        r.raise_for_status()
        deployments = r.json().get("deployments", [])
        if len(deployments):
            return {deployment["id"]: deployment for deployment in deployments}
    except Exception as e:
        logging.exception(e)
    return {}


def get_deployment(deployment_id: str, token) -> dict:
    return get_deployments(token)[deployment_id]


@ttl_cache(ttl=1)
def get_organizations(token):
    if not token:
        return {}

    headers = {"Authorization": f"Bearer {token}"}
    url = f"{ASTRO_ALPHA_API}/organizations"

    try:
        api_rv = requests.get(url, headers=headers).json()

        return {org["id"]: org for org in (api_rv or [])}
    except Exception as exc:
        print(exc)
        return {}


@ttl_cache(ttl=3600)
def get_jwk(token: str):
    jwks_url = f"{ASTRO_AUTH}/.well-known/jwks.json"
    jwks_client = jwt.PyJWKClient(jwks_url)
    return jwks_client.get_signing_key_from_jwt(token)


def get_deployment_url(deployment, token):
    all_astro_deployments = get_deployments(token)

    if all_astro_deployments and all_astro_deployments.get(deployment):
        url = urlparse(all_astro_deployments[deployment]["webServerUrl"])
        return f"https:/{url.netloc}/{url.path}"


def set_environment_variables(
    deployment: str, token: str, remote_vars: Dict[Dict[str, Any]]
):
    headers = {"Authorization": f"Bearer {token}"}
    client = GraphqlClient(endpoint=ASTROHUB_API, headers=headers)
    query = """
            fragment EnvironmentVariable on EnvironmentVariable {
                key
                value
                isSecret
                updatedAt
            }
            mutation deploymentVariablesUpdate($input: EnvironmentVariablesInput!) {
                deploymentVariablesUpdate(input: $input) {
                    ...EnvironmentVariable
                }
            }
            """
    client.execute(
        query,
        {
            "input": {
                "deploymentId": deployment,
                "environmentVariables": list(remote_vars.values()),
            }
        },
    )


def set_changed_environment_variables(
    deployment: str, token: str, items: Iterator[str]
) -> None:
    """
    :param deployment - deployment id
    :param token - user or deployment or workspace token
    :param items: str {'env-AIRFLOW__WEBSERVER__RELOAD_ON_PLUGIN_CHANGE': 'AIRFLOW__WEBSERVER__RELOAD_ON_PLUGIN_CHANGE'}
    """
    remote_vars = get_environment_variables(deployment, token)
    for item in items:
        if item not in remote_vars.keys():
            remote_vars.setdefault(
                item,
                {
                    "key": item,
                    "value": os.environ[item],
                    "isSecret": False,
                },
            )
    return set_environment_variables(deployment, token, remote_vars)


def get_environment_variables(deployment: str, token: str):
    try:
        # OLD STYLE? - part of the deployment
        return {
            remote_var["key"]: {
                "key": remote_var["key"],
                "value": remote_var["value"],
                "isSecret": remote_var["isSecret"],
            }
            for remote_var in get_deployment(deployment, token)["environmentVariables"]
        }
    except KeyError as e:
        logging.warning(e)
        logging.info("Getting ENV VARS via newer query")
        # NEW STYLE? - separate object
        headers = {
            "Authorization": f"Bearer {token}",
            "astro-current-org-id": "clacwluxo0u3i0t09epts2tjg",
        }
        client = GraphqlClient(
            #  api.astronomer.io/hub/graphql
            endpoint=ASTROHUB_GRAPHQL_API,
            headers=headers,
        )
        query = """
          fragment EnvironmentVariable on EnvironmentVariable {
            key
            value
            isSecret
            updatedAt
          }

          fragment DeploymentSpec on DeploymentSpec {
            environmentVariablesObjects {
              ...EnvironmentVariable
            }
          }

          query deploymentSpec($id: Id!) {
            deployment(id: $id) {
              id
              deploymentSpec {
                ...DeploymentSpec
              }
            }
          }"""
        r = client.execute(
            query,
            {"id": deployment},
        )
        return (
            r.get("data", {})
            .get("deploymentSpec", {})
            .get("environmentVariablesObjects", {})
        )


def is_environment_variable_migrated(deployment: str, token: str, key: str) -> bool:
    return key in get_environment_variables(deployment, token).keys()


@deprecated
class AstroClient:
    def __init__(self, token: Optional[str] = None, url: str = ASTROHUB_API):
        self.token = token
        self.url = url

    def list_deployments(self):
        headers = {"Authorization": f"Bearer {self.token}"}
        from python_graphql_client import GraphqlClient

        client = GraphqlClient(endpoint=self.url, headers=headers)
        query = """
                    {
                     deployments
                     {
                         id,
                         label,
                         releaseName,
                         workspace
                         {
                             id,
                             label
                         },
                         deploymentShortId,
                         deploymentSpec
                         {
                             environmentVariables
                             webserver {
                                 ingressHostname,
                                 url
                             }
                         }
                     }
                    }
                    """

        return client.execute(query).get("data")



import os

import requests
from airflow.models import BaseOperator, Connection
from airflow.utils.session import provide_session
from python_graphql_client import GraphqlClient

from astronomer.starship.variables.operators import AstroVariableMigrationOperator
from sqlalchemy.orm import Session
from typing import Any, Sequence


class AstroMigrationOperator(BaseOperator):
    """
    Sends connections, variables, and environment variables from a source Airflow to Astronomer Deployment
    """

    template_fields: Sequence[str] = ("token", "deployment_url")
    ui_color = "#974bde"

    def __init__(
        self,
        deployment_url,
        token,
        variables_exclude_list=None,
        connection_exclude_list=None,
        env_include_list=None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_url = deployment_url
        self.token = token
        self.variables_exclude_list = variables_exclude_list
        self.connection_exclude_list = connection_exclude_list
        self.env_include_list = env_include_list

    @provide_session
    def execute(self, context: Any, session: Session) -> None:
        variables = AstroVariableMigrationOperator(
            task_id="export_variables",
            deployment_url=self.deployment_url,
            token=self.token,
            variable_exclude_list=self.variables_exclude_list,
        )

        connections = AstroConnectionsMigrationOperator(
            task_id="export_connections",
            deployment_url=self.deployment_url,
            token=self.token,
            connection_exclude_list=self.connection_exclude_list,
        )

        env_vars = AstroEnvMigrationOperator(
            task_id="export_env_vars",
            deployment_url=self.deployment_url,
            token=self.token,
            env_include_list=self.env_include_list,
        )

        variables.execute(context=context)
        connections.execute(context=context)
        env_vars.execute(context=context)


class AstroConnectionsMigrationOperator(BaseOperator):
    """
    Sends connections from Airflow metadatabase to Astronomer Deployment
    """

    template_fields: Sequence[str] = ("token", "deployment_url")
    ui_color = "#974bde"

    def __init__(
        self, deployment_url, token, connection_exclude_list=None, **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_url = deployment_url
        self.token = token
        self.connection_exclude_list = connection_exclude_list

    @provide_session
    def execute(self, context: Any, session: Session) -> None:
        get_connections = session.query(Connection).all()
        local_connections = {conn.conn_id: conn for conn in get_connections}

        exclude_list = self.connection_exclude_list or []

        for key, value in local_connections.items():
            if key not in exclude_list:
                requests.post(
                    url=f"{self.deployment_url}/api/v1/connections",
                    headers={"Authorization": f"Bearer {self.token}"},
                    json={
                        "connection_id": key,
                        "conn_type": value.conn_type,
                        "host": value.host,
                        "login": value.login,
                        "schema": value.schema,
                        "port": value.port,
                        "password": value.password or "",
                        "extra": value.extra,
                    },
                )


class AstroEnvMigrationOperator(BaseOperator):
    """
    Sends env vars from Airflow to Astronomer Deployment
    """

    template_fields: Sequence[str] = ("token", "deployment_url")
    ui_color = "#974bde"

    def __init__(self, deployment_url, token, env_include_list, **kwargs) -> None:
        super().__init__(**kwargs)
        self.deployment_url = deployment_url
        self.token = token
        self.env_include_list = env_include_list

    def _find_deployment_id(self):
        deployments = self._astro_deployments()
        for id, deployment in deployments.items():
            url = deployment["deploymentSpec"]["webserver"]["url"]
            split_url = url.split("?", 1)
            base_url = split_url[0]
            if base_url in self.deployment_url:
                return id
                # TODO: Add logic if deployment id is not found

    def _existing_env_vars(self):
        deployment_id = self._find_deployment_id()

        client = GraphqlClient(
            endpoint="https://api.astronomer.io/hub/v1",
            headers={"Authorization": f"Bearer {self.token}"},
        )

        query = """
        {
            deployments
            {
                id,
                label,
                releaseName,
                workspace
                {
                    id,
                    label
                },
                deploymentShortId,
                deploymentSpec
                {
                    environmentVariables
                    webserver {
                        ingressHostname,
                        url
                    }
                }
            }
        }
        """

        try:
            env_vars = {}
            deployments = client.execute(query,)[
                "data"
            ]["deployments"]
            for deployment in deployments:
                if deployment["id"] == deployment_id:
                    env_vars = deployment["deploymentSpec"]["environmentVariables"]

            existing_vars = {}
            for var in env_vars:
                existing_vars[var["key"]] = {
                    "key": var["key"],
                    "value": var["value"],
                    "isSecret": var["isSecret"],
                }

            return existing_vars
        except Exception as exc:
            print(exc)
            return {}

    def _astro_deployments(self):
        headers = {"Authorization": f"Bearer {self.token}"}
        client = GraphqlClient(
            endpoint="https://api.astronomer.io/hub/v1", headers=headers
        )
        query = """
        {
            deployments
            {
                id,
                deploymentSpec
                {
                    webserver {
                        url
                    }
                }
            }
        }
        """

        try:
            api_rv = client.execute(query)["data"]["deployments"]

            return {deploy["id"]: deploy for deploy in (api_rv or [])}
        except Exception as exc:
            print(exc)
            return {}

    @provide_session
    def execute(self, context: Any, session: Session) -> None:
        client = GraphqlClient(
            endpoint="https://api.astronomer.io/hub/v1",
            headers={"Authorization": f"Bearer {self.token}"},
        )
        query = """
        fragment EnvironmentVariable on EnvironmentVariable {
            key
            value
            isSecret
            updatedAt
        }
        mutation deploymentVariablesUpdate($input: EnvironmentVariablesInput!) {
            deploymentVariablesUpdate(input: $input) {
                ...EnvironmentVariable
            }
        }
        """
        deployment = self._find_deployment_id()

        complete_env_list = self._existing_env_vars()

        for key, value in os.environ.items():
            if self.env_include_list:
                if key in self.env_include_list and self.env_include_list:
                    complete_env_list[key] = {
                        "key": key,
                        "value": value,
                        "isSecret": False,
                    }

        client.execute(
            query,
            {
                "input": {
                    "deploymentId": deployment,
                    "environmentVariables": list(complete_env_list.values()),
                }
            },
        )



import datetime
import json
import logging
import os
import socket
import urllib
from contextlib import redirect_stderr, redirect_stdout
from typing import Sequence

from airflow.exceptions import AirflowFailException
from airflow.models import BaseOperator

from astronomer.aeroscope.util import clean_airflow_report_output


class AeroscopeOperator(BaseOperator):
    template_fields: Sequence[str] = ("presigned_url", "organization")

    def __init__(
        self,
        presigned_url,
        organization,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.presigned_url = presigned_url
        self.organization = organization

    def execute(self, context):
        import io
        import runpy
        from urllib.request import urlretrieve

        import requests

        VERSION = os.getenv("TELESCOPE_REPORT_RELEASE_VERSION", "latest")
        a = "airflow_report.pyz"
        if VERSION == "latest":
            self.log.info("Running Latest Version of report")
            urlretrieve(
                "https://github.com/astronomer/telescope/releases/latest/download/airflow_report.pyz",
                a,
            )
        else:
            try:
                self.log.info(f"Running Version {VERSION} of report")
                urlretrieve(
                    f"https://github.com/astronomer/telescope/releases/download/{VERSION}/airflow_report.pyz",
                    a,
                )
            except urllib.error.HTTPError as e:
                raise AirflowFailException(
                    f"Error finding specified version:{VERSION} -- Reason:{e.reason}"
                )
        s = io.StringIO()
        with redirect_stdout(s), redirect_stderr(s):
            runpy.run_path(a)
        date = datetime.datetime.now(datetime.timezone.utc).isoformat()[:10]
        content = {
            "telescope_version": "aeroscope-latest",
            "report_date": date,
            "organization_name": self.organization,
            "local": {
                socket.gethostname(): {
                    "airflow_report": clean_airflow_report_output(s.getvalue())
                }
            },
        }
        upload = requests.put(self.presigned_url, data=json.dumps(content))
        if not upload.ok:
            logging.error(
                f"Upload failed with code {upload.status_code}::{upload.text}"
            )
            upload.raise_for_status()
